from typing import TypeVar, Callable, Any

T = TypeVar('T')

def feed_to_fluent_callable(obj, collector: T) -> T:
    field_name = 'exec'
    existing = None
    if hasattr(collector,field_name):
        existing = getattr(collector,field_name)
        if not isinstance(existing,DelayedExecution):
            if existing is not None:
                raise ValueError("Collector instance already has attribute `{0}`, and it's not of type DelayedExecution. Specify another name for the finalizing field".format(field_name))
        else:
            if existing.initialized:
                raise ValueError('Collector was already initialized.')
            existing._initialize(collector,obj)

    if existing is None:
        ex = DelayedExecution()
        ex._initialize(collector,obj)
        setattr(collector,field_name,ex)
    return collector

class DelayedExecution:
    def __init__(self):
        self.call = None
        self.argument = None
        self.initialized = False

    def _initialize(self, call, argument):
        self.call = call
        self.argument = argument
        self.initialized = True

    def _check_initialization(self):
        if not self.initialized:
            raise ValueError('DelayedExecution was not initialized')

    def result(self):
        self._check_initialization()
        return self.call(self.argument)

    def feed(self, next_call: Callable[[Any],T]) -> T:
        self._check_initialization()
        return next_call(self.call(self.argument))

    def feed_to_fluent_callable(self, next_collector: T) -> T:
        self._check_initialization()
        return feed_to_fluent_callable(self.call(self.argument), next_collector)






from unittest import TestCase
from yo_fluq import *

class AllAnyTest(TestCase):

    def test_all_1(self):
        self.assertEqual(True, Query.args(1, 2, 3).all(lambda z: z > 0))

    def test_all_2(self):
        self.assertEqual(False, Query.args(1, 2, 3).all(lambda z: z > 1))

    def test_all_3(self):
        self.assertEqual(True, Query.args(dict()).all())

    def test_all_4(self):
        self.assertEqual(True, Query.args().all())


    def test_any_1(self):
        self.assertEqual(True, Query.args(1, 2, 3).any(lambda z: z > 1))

    def test_any_2(self):
        self.assertEqual(False, Query.args(1, 2, 3).any(lambda z: z < 1))

    def test_any_3(self):
        self.assertEqual(True, Query.args(dict()).any())

    def test_any_4(self):
        self.assertEqual(False, Query.args().any())




from .._common import *
from .._push_queries import AggregationCodeFactory, agg
from .queryable_code_factory import QueryableCodeFactory


class Queryable(Generic[T],AggregationCodeFactory,QueryableCodeFactory['Queryable',T]):
    def __init__(self,en,length=None):
        AggregationCodeFactory.__init__(self,self._aggregate_with)
        QueryableCodeFactory.__init__(self,en,length)

    def _aggregate_with(self, aggregator: agg.PushQueryElement):
        return aggregator(self.en)





from yo_fluq import *
from unittest import TestCase

class TestFeed(TestCase):
    def test_feed_execute(self):
        q = (Query
             .args(1,2,3,4)
             .feed_to_fluent_callable(Query.push())
             .split_by_group(lambda z: z%2)
             .sum()
             .exec.result()
             )
        self.assertDictEqual({0:6,1:4}, q)

    def test_feed_feed(self):
        q = (Query
             .args(1,2,3,4)
             .feed_to_fluent_callable(Query.push())
             .split_by_group(lambda z: z%2)
             .sum()
             .exec.feed(set)
        )
        self.assertSetEqual(
            {0,1},
            q
        )

    def test_feed_feed_fluent(self):
        q = (Query
             .args(1,2,3,4)
             .feed_to_fluent_callable(Query.push())
             .split_by_group(lambda z: z % 2)
             .sum()
             .exec.feed(Query.dict)
             .order_by(lambda z: z.key)
             .select(lambda z: z.value)
             .to_list()
        )
        self.assertListEqual(
            [6,4],
            q
        )

    def test_feed_feed_fluent_callable(self):
        q = (
            Query
            .args(1,2,3, 4)
            .feed_to_fluent_callable(Query.push())
            .split_by_group(lambda z: z%2)
            .sum()
            .exec.feed_to_fluent_callable(Query.push())
            .count()
            .exec.result()
        )
        self.assertEqual(2,q)

